/* hdparm.c - Command line interface to get/set hard disk parameters */
/*          - by Mark S. Lord (c) 1994 -- freely distributable */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <linux/hdreg.h>
#include <linux/fs.h>
#include <linux/major.h>

#ifndef HD1_MAJOR
#define HD1_MAJOR 22	/* for second controller using atdisk2 patches */
#endif

#define TIMING_MB		16
#define TIMING_BUF_MB		1
#define TIMING_BUF_BYTES	(TIMING_BUF_MB * 1024 * 1024)
#define TIMING_BUF_COUNT	(TIMING_MB / TIMING_BUF_MB)
#define BUFCACHE_FACTOR		2

char *progname;
struct hd_driveid id;
static long unmask = 0, readonly = 0, mult = 0, readahead = 0;
static int verbose = 0, get_identity = 0, get_geom = 0, get_size = 0;
static int set_unmask = 0, set_readonly = 0, set_mult = 0, set_readahead = 0;
static int get_unmask = 0, get_readonly = 0, get_mult = 0, get_readahead = 0;
static int flagcount  = 0, do_timings = 0, is_scsi_hd = 0;

char *cfg_str[] =
{	"", " HardSect", " SoftSect", " NotMFM", " HdSw>15uSec", " SpinMotCtl",
	" Fixed", " Removeable", " DTR<=5Mbs", " DTR>5Mbs", " DTR>10Mbs",
	" RotSpdTol>.5%", " dStbOff", " TrkOff", " FmtGapReq", " nonMagnetic"
};

char *SlowMedFast[]	= {"0(slow)", "1(medium)", "2(fast)"};
char *BuffType[]	= {"?", "1Sect", "DualPort", "DualPortCache"};

#define YN(b)	(((b)==0)?"no":"yes")

static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
{
	printf (prefix);
	printf ( (i > maxi) ? "?" : s[i] );
}

static void dump_identity (struct hd_driveid *id)
{
	int i;
	printf ("\n Model=%.40s, FwRev=%.20s, SerialNo=%.8s", 
		id->model, id->fw_rev, id->serial_no);
	printf ("\n Config={");
	for (i=0; i<=15; i++)
		if (id->config & (1<<i)) printf ("%s",cfg_str[i]);
	printf (" }\n");
	printf (" Physical CHS=%d/%d/%d, TrkSize=%d, SectSize=%d, ECCbytes=%d\n",
		id->cyls, id->heads, id->sectors,
		id->track_bytes, id->sector_bytes, id->ecc_bytes);
	dmpstr (" BuffType=",id->buf_type,BuffType,3);
	printf (", BuffSize=%dKB, MaxMultSect=%d\n", id->buf_size/2, id->max_multsect);
	printf (" Features: DblWordIO=%s, LBA=%s, DMA=%s",
		YN(id->dword_io&1),YN(id->capability&2),YN(id->capability&1));
	dmpstr (", tPIO=",id->tPIO,SlowMedFast,2);
	if (id->capability&1)
		dmpstr (", tDMA=",id->tDMA,SlowMedFast,2);
	if ((id->capability&8) || (id->cur_valid&2)) {
		printf ("\nEIDE: ");
		if (id->capability&8) {
			if (id->capability&4)
				printf ("IORDY=switchable");
			else
				printf ("IORDY=yes");
		}
		if (id->cur_valid&2) {
			if (id->capability&8)
				printf(", ");
			printf ("Mode3=yes");
		}
	}
	printf ("\n (%s): Logical CHS=%d/%d/%d, TotSect=%d",
		(id->cur_valid&1)?"valid":"maybe",
		id->cur_cyls, id->cur_heads, id->cur_sectors, *(int *)&id->cur_capacity0);
	if (id->capability&2)
		printf (", MaxLBAsect=%d", id->lba_capacity);
	printf ("\n MultSect=");
	if (!id->max_multsect)
		printf ("n/a");
	else if (!(id->multsect_valid&1))
		printf ("???");
	else if (id->multsect)
		printf ("%d", id->multsect);
	else
		printf ("off");
	if (id->capability&1)
		printf (", DMA-1w=%04X, DMA-mw=%04X", id->dma_1word, id->dma_mword);
	printf ("\n\n");
}

void flush_buffer_cache (int fd)
{
	if (ioctl(fd, BLKFLSBUF, NULL)) {
		perror("BLKFLSBUF failed");
		return;
	}
}

int seek_to_zero (int fd)
{
	if (lseek(fd, (off_t) 0, SEEK_SET)) {
		perror("lseek() failed");
		return 1;
	}
	return 0;
}

int read_big_block (int fd, char *buf)
{
	int rc;
	if ((rc = read(fd, buf, TIMING_BUF_BYTES)) != TIMING_BUF_BYTES) {
		if (rc)
			perror("read() failed");
		else
			fputs("read() hit EOF - device too small\n", stderr);
		return 1;
	}
	return 0;
}

void time_device (int fd)
{
	int i;
	char *buf;
	unsigned long cpu;
	double elapsed, correction;
	struct itimerval e1, e2;
	struct tms c1, c2;
	int shmid;

	if ((shmid = shmget(IPC_PRIVATE, TIMING_BUF_BYTES, 0600)) == -1) {
		perror ("could not allocate sharedmem buf");
		return;
	}
	if (shmctl(shmid, SHM_LOCK, NULL) == -1) {
		perror ("could not lock sharedmem buf");
		(void) shmctl(shmid, IPC_RMID, NULL);
		return;
	}
	if ((buf = shmat(shmid, (char *) 0, 0)) == (char *) -1) {
		perror ("could not attach sharedmem buf");
		(void) shmctl(shmid, IPC_RMID, NULL);
		return;
	}
	if (shmctl(shmid, IPC_RMID, NULL) == -1)
		perror ("shmctl(,IPC_RMID,) failed");

	/* Clear out the device request queues & give them time to complete */
	sync();
	sleep(3);

	/* Calculate a correction factor for the basic
	 * overhead of doing a read() from the buffer cache.
	 * To do this, we read the data once to "cache it" and
	 * to force full preallocation of our timing buffer,
	 * and then we re-read it 10 times while timing it.
	 *
	 * getitimer() is used rather than gettimeofday() because
	 * it is much more consistent (on my machine, at least).
	 */
	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
	if (seek_to_zero (fd)) return;
	if (read_big_block (fd, buf)) return;
	/* access all sectors of buf to ensure the read fully completed */
	for (i = 0; i < TIMING_BUF_BYTES; i += 512)
		buf[i] &= 1;
	printf (" Timing buffer-cache reads:   ");
	fflush(stdout);

	/* Clear out the device request queues & give them time to complete */
	sync();
	sleep(1);

	/* Time re-reading from the buffer-cache */
	getitimer(ITIMER_REAL, &e1);
	for (i = (BUFCACHE_FACTOR * TIMING_BUF_COUNT) ; i > 0; --i) {
		if (seek_to_zero (fd)) goto quit;
		if (read_big_block (fd, buf)) goto quit;
	}
	getitimer(ITIMER_REAL, &e2);
	correction = (e1.it_value.tv_sec - e2.it_value.tv_sec)
	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);

	/* Now remove the lseek() from the correction factor */
	getitimer(ITIMER_REAL, &e1);
	for (i = (BUFCACHE_FACTOR * TIMING_BUF_COUNT) ; i > 0; --i) {
		if (seek_to_zero (fd)) goto quit;
	}
	getitimer(ITIMER_REAL, &e2);
	correction -= (e1.it_value.tv_sec - e2.it_value.tv_sec)
	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);

	printf ("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
		(BUFCACHE_FACTOR * TIMING_MB), correction,
		(BUFCACHE_FACTOR * TIMING_MB) / correction);
	correction /= BUFCACHE_FACTOR;

	flush_buffer_cache(fd);
	sleep(1);

	printf (" Timing buffered disk reads:  ");
	fflush(stdout);

	/* Now do the timings for real */
	cpu = times(&c1);
	getitimer(ITIMER_REAL, &e1);
	for (i = TIMING_BUF_COUNT; i > 0; --i) {
		if (read_big_block (fd, buf)) goto quit;
	}
	getitimer(ITIMER_REAL, &e2);
	cpu = times(&c2) - cpu;
	cpu = ((c2.tms_utime + c2.tms_stime) - (c1.tms_utime + c1.tms_stime))
		 * 100 / cpu;

	elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
	printf ("%2d MB in %5.2f seconds =%5.2f MB/sec (%ld%% CPU)\n",
		TIMING_MB, elapsed, TIMING_MB / elapsed, cpu);

	if (elapsed <= (correction * 2))
		printf("Hmm.. suspicious results: probably not enough memory for a proper test.\n");
	else {
		printf (" Estimating raw device speed: ");
		elapsed -= correction;
		printf ("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
			TIMING_MB, elapsed, TIMING_MB / elapsed);
	}
quit:
	if (-1 == shmdt(buf))
		perror ("could not detach sharedmem buf");
}

void no_scsi (void)
{
	if (is_scsi_hd) {
		fputs(" operation not supported on SCSI disks\n", stderr);
		exit(EINVAL);
	}
}

void process_dev (char *devname)
{
	int fd, rc;
	static long parm, multcount;
	struct stat stat_buf;

	if (stat(devname,&stat_buf)) {
		perror(devname);
		exit(errno);
	}

	is_scsi_hd = (major(stat_buf.st_rdev) == SCSI_DISK_MAJOR);

	if (major(stat_buf.st_rdev) != HD_MAJOR
	 && major(stat_buf.st_rdev) != HD1_MAJOR && !is_scsi_hd) {
		fprintf (stderr,"%s is not a hard disk.\n",devname);
		exit(EINVAL);
	}

	fd = open (devname, O_RDONLY);
	if (fd < 0) {
		perror(devname);
		exit(errno);
	}

	printf ("\n%s:\n", devname);
	if (set_readahead) {
		printf (" setting fs readahead to %ld\n", readahead);
		if (ioctl(fd, BLKRASET, readahead)) 
			perror(" BLKRASET failed");
	}
	if (set_mult) {
		no_scsi();
		printf (" setting multcount to %ld\n", mult);
		if (ioctl(fd, HDIO_SETMULTCOUNT, &mult)) 
			perror(" HDIO_SETMULTCOUNT failed");
		else {
			char buf[512];
			if ((rc = ioctl(fd, BLKFLSBUF, NULL))) 
				perror(" BLKFLSBUF failed");
			if (-1 == read(fd, buf, sizeof(buf)))
				perror(" read failed");
		}
	}
	if (set_readonly) {
		printf (" setting readonly to %ld\n", readonly);
		if (ioctl(fd, BLKROSET, &readonly))
			perror(" BLKROSET failed");
	}
	if (set_unmask) {
		no_scsi();
		printf (" setting unmaskirq to %ld\n", unmask);
		if (ioctl(fd, HDIO_SETUNMASKINTR, &unmask))
			perror(" HDIO_SETUNMASKINTR failed");
	}

	if (!flagcount)
		verbose = 1;

	if (verbose || get_readahead) {
		if (ioctl(fd, BLKRAGET, &parm))
			perror(" BLKRAGET failed");
		else    printf (" readahead = %ld\n", parm);
	}
	if ((verbose && !is_scsi_hd) || get_mult || get_identity) {
		no_scsi();
		multcount = -1;
		if (ioctl(fd, HDIO_GETMULTCOUNT, &multcount))
			perror(" HDIO_GETMULTCOUNT failed");
		else if (get_mult)
			printf (" multcount = %ld\n", multcount);
	}
	if (verbose || get_readonly) {
		if (ioctl(fd, BLKROGET, &parm))
			perror(" BLKROGET failed");
		else    printf (" readonly  = %ld\n", parm);
	}
	if ((verbose && !is_scsi_hd) || get_unmask) {
		no_scsi();
		if (ioctl(fd, HDIO_GETUNMASKINTR, &parm))
			perror(" HDIO_GETUNMASKINTR failed");
		else    printf (" unmaskirq = %ld\n", parm);
	}
	if (verbose || get_geom) {
		static struct hd_geometry g;
		if (ioctl(fd, HDIO_GETGEO, &g))
			perror(" HDIO_GETGEO failed");
		else    printf (" geometry  = %d/%d/%d, offset = %ld\n",
				g.cylinders, g.heads, g.sectors, g.start);
	}
	if (verbose || get_size) {
		if (ioctl(fd, BLKGETSIZE, &parm))
			perror(" BLKGETSIZE failed");
		else    printf (" sectors   = %ld\n", parm);
	}

	if (get_identity) {
		no_scsi();
		if (!(ioctl(fd, HDIO_GETIDENTITY, &id))) {
			if (multcount != -1) {
				id.multsect = multcount;
				id.multsect_valid |= 1;
			} else
				id.multsect_valid &= ~1;
			dump_identity(&id);
		} else if (errno == -ENOMSG)
			printf (" no identification info available\n");
		else
			perror(" HDIO_GETIDENTITY failed");
	}
	if (do_timings)
		time_device (fd);
	close (fd);
}

void usage_error (void)
{
	fprintf (stderr,"%s - get/set hard disk parameters - version 1.4\n\n", progname);
	fprintf (stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
	fprintf (stderr,"Options:\n");
	fprintf (stderr," -a	get/set fs readahead\n");
	fprintf (stderr," -g	display drive geometry\n");
	fprintf (stderr," -h	display terse usage information\n");
	fprintf (stderr," -i  *	display drive identification\n");
	fprintf (stderr," -m  *	get/set multiple sector count\n");
	fprintf (stderr," -r	get/set readonly flag\n");
	fprintf (stderr," -s	display device size in sectors\n");
	fprintf (stderr," -t	perform disk read timings\n");
	fprintf (stderr," -u  *	get/set unmaskirq flag\n");
	fprintf (stderr," -v	same as -agmrsu (default)\n\n");
	fprintf (stderr,"	* only for IDE drives\n\n");
	exit(1);
}

int main(int argc, char **argv)
{
	char c, *p;

	if  ((progname = (char *) strrchr(*argv, '/')) == NULL)
                progname = *argv;
        else
                progname++;
	++argv;

	if (!--argc)
		usage_error();
	while (argc--) {
		p = *argv++;
		if (*p == '-') {
			if (!*++p)
				usage_error();
			while ((c = *p++)) {
				++flagcount;
				switch (c) {
					case 'v':
						verbose = 1;
						break;
					case 'i':
						get_identity = 1;
						break;
					case 'g':
						get_geom = 1;
						break;
					case 's':
						get_size = 1;
						break;
					case 'u':
						get_unmask = 1;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						if (*p == '0' || *p == '1') {
							set_unmask = 1;
							unmask = *p++ - '0';
						}
						break;
					case 'r':
						get_readonly = 1;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						if (*p == '0' || *p == '1') {
							set_readonly = 1;
							readonly = *p++ - '0';
						}
						break;
					case 'm':
						get_mult = 1;
						mult = 0;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						while (isdigit(*p)) {
							set_mult = 1;
							mult *= 10;
							mult += *p++ - '0';
						}
						break;
					case 'a':
						get_readahead = 1;
						readahead = 0;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						while (isdigit(*p)) {
							set_readahead = 1;
							readahead *= 10;
							readahead += *p++ - '0';
						}
						break;
					case 't':
						do_timings = 1;
						break;
					case 'h':
					default:
						usage_error();
				}
			}
			if (!argc)
				usage_error();
		} else {
			process_dev (p);
		} 
	}
	exit (0);
}
